
"""

"""

from tingyun.packages import six
from tingyun.armoury.ammunition.external_tracker import wrap_external_trace


def detect(module):
    def url_opener_open(instance, fullurl, *args, **kwargs):
        """
        :param instance:
        :param fullurl:
        :param args:
        :param kwargs:
        :return:
        """
        if isinstance(fullurl, six.string_types):
            return fullurl
        else:
            return fullurl.get_full_url()

    wrap_external_trace(module, 'OpenerDirector.open', 'urllib2', url_opener_open)


from collections import namedtuple


_ErrorNode = namedtuple('_ErrorNode', ['error_time', 'http_status', "error_class_name", 'uri', 'thread_name',
                                       "message", 'stack_trace', 'request_params', "tracker_type", "referer"])


class ErrorNode(_ErrorNode):
    """

    """
    pass




""" define some logistics wrapper

"""

import logging

from tingyun.packages.wrapt.wrappers import (FunctionWrapper as _FunctionWrapper)
from tingyun.packages.wrapt.wrappers import BoundFunctionWrapper as _BoundFunctionWrapper

from tingyun.packages.wrapt.wrappers import wrap_object as _wrap_object, wrap_function_wrapper as _wrap_function_wrapper
from tingyun.packages.wrapt.wrappers import function_wrapper as _function_wrapper

console = logging.getLogger(__name__)


# ----------------------------------------------------------------------------
# following method are used the wrapt module instead.
# all methods would send from this module, not use the wrapt package directly

# egg: (memcached, Client.append, MemcacheTraceWrapper(), get)
wrap_object = _wrap_object

# for wrap the function(bound/unbound)
FunctionWrapper = _FunctionWrapper
BoundFunctionWrapper = _BoundFunctionWrapper

wrap_function_wrapper = _wrap_function_wrapper

function_wrapper = _function_wrapper



"""define tracker for tracer use.

"""

import logging
import time
import sys
import traceback

from tingyun.armoury.ammunition.timer import Timer
from tingyun.battlefield.knapsack import knapsack
from tingyun.logistics.warehouse.database_node import DatabaseNode
from tingyun.logistics.warehouse.tracker_node import TrackerNode
from tingyun.logistics.warehouse.error_node import ErrorNode
from tingyun.packages import six

console = logging.getLogger(__name__)


class Tracker(object):
    """Tracker, the battle entrance.
    """
    def __init__(self, proxy, enabled=None, framework="Python"):
        """
        :param proxy: communication proxy instance to controller.
        :return:
        """
        self.framework = framework
        self.proxy = proxy
        self.enabled = False

        self.thread_id = knapsack().current_thread_id()
        self._tracker_id = id(self)

        self.background_task = False
        self.start_time = 0
        self.end_time = 0
        self.trace_node = []

        self._errors = []
        self._custom_params = {}
        self._slow_sql_nodes = []

        self.request_uri = None
        self.http_status = 0
        self.request_params = {}
        self._priority = 0
        self._group = None
        self._name = None
        self.apdex = 0
        self._frozen_path = None

        self.stack_trace_count = 0
        self.explain_plan_count = 0

        self.thread_name = "Unknown"
        self.referer = ""

        # set the global switch of agent.  the return application settings is download from server witch is merged
        # with the local settings on server
        # agent always be enabled if enabled in local settings. just not upload data when disabled by server.
        global_settings = proxy.global_settings
        if global_settings.enabled:
            if enabled or (enabled is None and proxy.enabled):
                self._settings = proxy.settings
                if not self._settings:
                    self.proxy.activate()
                    self._settings = proxy.settings

                if self._settings:
                    self.enabled = True

    def __enter__(self):
        """
        :return:
        """
        if not self.enabled:
            return self

        try:
            self.save_tracker()
        except Exception as _:
            console.fatal("Fatal error when save tracker. if this continues. please contact us for further \
                           investigation")
            raise

        self.start_time = time.time()
        self.trace_node.append(Timer(None))

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        if not self.enabled:
            return

        try:
            self.drop_tracker()
        except Exception as err:
            console.fatal("Agent suffered fatal error when drop tracker. please contact us for further \
                           investigation")
            console.exception("error detail %s", err)
            raise

        if not self.is_uri_captured(self.request_uri):
            console.debug("ignore the uri %s", self.request_uri)
            return

        if exc_type is not None and exc_val is not None and exc_tb is not None:
            self.record_exception(exc_type, exc_val, exc_tb)

        self.end_time = time.time()
        duration = int((self.end_time - self.start_time) * 1000)
        root_node = self.trace_node.pop()
        children = root_node.children
        exclusive = duration + root_node.exclusive

        tracker_type = "WebAction" if not self.background_task else "BackgroundAction"
        group = self._group or ("Python" if self.background_task else "Uri")
        request_params = self.filter_params(self.request_params)
        uri = self.url_encode(self.request_uri)

        # replace the specified metric apdex_t
        apdex_t = self._settings.apdex_t
        path = self.path
        if path in self._settings.action_apdex:
            apdex_t = self._settings.action_apdex.get(path)

        node = TrackerNode(type=tracker_type, group=group, name=self._name, start_time=self.start_time,
                           end_time=self.end_time, request_uri=uri, duration=duration, thread_name=self.thread_name,
                           http_status=self.http_status, exclusive=exclusive, children=tuple(children),
                           path=self.path, errors=self._errors, apdex_t=apdex_t,
                           custom_params=self._custom_params, request_params=request_params,
                           referer=self.referer, slow_sql=self._slow_sql_nodes)

        self.proxy.record_tracker(node)

    def record_exception(self, exc=None, value=None, tb=None, params=None, tracker_type="WebAction",
                         ignore_errors=[]):
        """ record the exception for trace
        :param exc:
        :return:
        """
        if not self._settings.error_collector.enabled:
            return

        if exc is None and value is None and tb is None:
            exc, value, tb = sys.exc_info()

        if exc is None or value is None or tb is None:
            console.warning("None exception is got. skip it now. %s, %s, %s", exc, value, tb)
            return

        if self.http_status in self._settings.error_collector.ignored_status_codes:
            console.debug("ignore the status code %s", self.http_status)
            return

        # 'True' - ignore the error.
        # 'False'- record the error.
        # ignore status code and maximum error number filter will be done in data engine because of voiding repeat count
        # method ignore_errors() is used to detect the the status code which is can not captured
        if callable(ignore_errors):
            should_ignore = ignore_errors(exc, value, tb, self._settings.error_collector.ignored_status_codes)
            if should_ignore:
                return

        # think more about error occurred before deal the status code.
        if self.http_status in self._settings.error_collector.ignored_status_codes:
            console.debug("record_exception: ignore  error collector status code")
            return

        module = value.__class__.__module__
        name = value.__class__.__name__
        fullname = '%s:%s' % (module, name) if module else name

        request_params = self.filter_params(self.request_params)
        if params:
            custom_params = dict(request_params)
            custom_params.update(params)
        else:
            custom_params = dict(request_params)

        try:
            message = str(value)
        except Exception as _:
            try:
                # Assume JSON encoding can handle unicode.
                message = six.text_type(value)
            except Exception as _:
                message = '<unprintable %s object>' % type(value).__name__

        stack_trace = traceback.extract_tb(tb)
        node = ErrorNode(error_time=int(time.time()), http_status=self.http_status, error_class_name=fullname,
                         uri=self.url_encode(self.request_uri), thread_name=self.thread_name, message=message,
                         stack_trace=stack_trace, request_params=custom_params, tracker_type=tracker_type,
                         referer=self.referer)

        self._errors.append(node)

    @property
    def path(self):
        """the call tree path
        """
        if self._frozen_path:
            return self._frozen_path

        tracker_type = "WebAction" if not self.background_task else "BackgroundAction"
        name = self._name or "Undefined"

        uri_params_captured = self._settings.web_action_uri_params_captured.get(self.request_uri, '')
        if self.request_uri in self._settings.web_action_uri_params_captured:
            match_param = ''
            for actual_param in self.request_params:
                if actual_param in uri_params_captured:
                    match_param += "&%s=%s" % (actual_param, self.request_params[actual_param])
            match_param = match_param.replace("&", "?", 1)
            self.request_uri = "%s%s" % (self.request_uri, match_param)

        if not self._settings.auto_action_naming:
            path = "%s/%s/%s" % (tracker_type, self.framework, self.url_encode(self.request_uri))
        else:
            path = '%s/%s/%s' % (tracker_type, self.framework, name)

        return path

    def url_encode(self, uri):
        """
        :param uri:
        :return:
        """
        encoded_url = "index"
        if not uri or uri == "/":
            return encoded_url

        # drop the uri first /
        uri = uri.replace('/', '', 1)
        encoded_url = uri.replace("/", "%2F")

        return encoded_url

    def set_tracker_name(self, name, group="Function", priority=None):
        """used to consist metric identification
        """
        if self._priority is None:
            return

        if priority is not None and priority < self._priority:
            return

        if isinstance(name, bytes):
            name = name.decode('Latin-1')

        self._priority = priority
        self._group = group
        self._name = name

    def save_tracker(self):
        knapsack().save_tracker(self)

    def drop_tracker(self):
        knapsack().drop_tracker(self)

    def process_database_node(self, node):
        """record the database node
        :param node: database node
        :return:
        """
        if type(node) is not DatabaseNode:
            return

        if not self._settings.action_tracer.enabled:
            return

        if not self._settings.action_tracer.slow_sql:
            return

        if self._settings.action_tracer.record_sql == 'off':
            return

        if node.duration < self._settings.action_tracer.slow_sql_threshold:
            return

        self._slow_sql_nodes.append(node)

    def push_node(self, node):
        self.trace_node.append(node)

    def pop_node(self, node):
        """return parent trace node
        """
        last = self.trace_node.pop()
        assert last == node
        parent = self.trace_node[-1]

        return parent

    def current_node(self):
        """
        """
        if self.trace_node:
            return self.trace_node[-1]

    def is_uri_captured(self, uri):
        """check the uri is captured or not
        :param uri: uri of the request
        :return: True, if it's need to captured. else return False
        """
        # capture all url
        if not self._settings.urls_captured:
            return True

        for p in self._settings.urls_captured:
            if p and p.match(uri):
                return True

        return False

    def filter_params(self, params):
        """filter the parameters
        :param params: the dict parameters
        :return: filtered parameters
        """
        result = {}

        if not params:
            return result

        if not self._settings.action_tracer.enabled or not self._settings.capture_params:
            return result

        for key in params:
            if key not in self._settings.ignored_params:
                result[key] = params[key]

        return result

    @property
    def name(self):
        return self._name

    @property
    def group(self):
        return self._group

    @property
    def settings(self):
        return self._settings


def current_tracker():
    """return current tracker in the thread
    :return:
    """
    return knapsack().current_tracker()




"""this module is defined to wrap the thrift.
"""

from tingyun.armoury.ammunition.external_tracker import wrap_external_trace


def detect_tsocket(module):
    """
    :param module:
    :return:
    """
    def tsocket_url(tsocket, *args, **kwargs):
        # _unix_socket is passed in when TSocket initialized
        # when wrap the specified func, the tsocket is instance.
        url = 'thrift://%s:%s' % (tsocket.host, tsocket.port)

        if tsocket._unix_socket is None:
            url = 'thrift://%s:%s' % (tsocket.host, tsocket.port)

        return url

    wrap_external_trace(module, 'TSocket.open', 'thrift', tsocket_url)


def detect_tsslsocket(module):
    """
    :param module:
    :return:
    """
    def tsocket_ssl_url(tsocket, *args, **kwargs):
        # _unix_socket is passed in when TSSSocket initialized
        # when wrap the specified func, the TSSSocket is instance.
        url = 'thrift//%s:%s' % (tsocket.host, tsocket.port)

        if tsocket._unix_socket is None:
            url = 'thrift//%s:%s' % (tsocket.host, tsocket.port)

        return url

    wrap_external_trace(module, 'TSSLSocket.open', 'thrift', tsocket_ssl_url)




"""transportation base tool for upload

"""

import time
import logging
import sys
import json
import zlib
from tingyun.logistics.exceptions import DiscardDataForRequest, RetryDataForRequest, InvalidLicenseException
from tingyun.logistics.exceptions import InvalidDataTokenException, OutOfDateConfigException, ServerIsUnavailable
from tingyun.config.settings import global_settings
from tingyun.packages import six, requests

console = logging.getLogger(__name__)

# 10kbi
COMPRESS_MINIMUM_SIZE = 10 * 1024
USER_AGENT = ' NBS Newlens Agent/%s (python %s; %s)' % (global_settings().agent_version, sys.version.split()[0],
                                                        sys.platform)


def transmitter(session, url, action, payload={}, param={}, audit_mode=False):
    """
    :param session: the request session to server
    :param url: the address witch data send to
    :param action: the send actually intention
    :param payload: request data
    :return: None
    """
    console.debug("Send request with url %s, action %s param %s", url, action, param)
    settings = global_settings()
    start_time = time.time()
    headers = {
        "User-Agent": USER_AGENT,
        "Content-Type": "application/octet-stream",
        "connection": "close",
        "Accept-Encoding": "deflate",
    }

    try:
        data = json.dumps(payload)
    except Exception as err:
        console.error("Encoding json for payload failed, url %s action %s param %s payload %s err %s",
                      url, action, param, payload, err)
        raise DiscardDataForRequest(str(sys.exc_info()[1]))

    if len(data) > COMPRESS_MINIMUM_SIZE:
        headers['Content-Encoding'] = 'deflate'
        level = (len(data) < 2000000) and 1 or 9
        data = zlib.compress(six.b(data), level)

    auto_close_session = False
    if not session:
        session = requests.session()
        auto_close_session = True

    content = ""
    try:
        timeout = settings.data_report_timeout
        verify_ssl = False if settings.port else True
        ret = session.post(url, data=data, params=param, headers=headers, timeout=timeout, verify=verify_ssl)
        content = ret.content
    except requests.RequestException:
        console.error('Agent server is not attachable. if the error continues, please report to networkbench.'
                      'The error raised was %s. response content is %s.', sys.exc_info()[1], content)
        raise RetryDataForRequest(str(sys.exc_info()[1]))
    finally:
        duration = time.time() - start_time
        if auto_close_session:
            session.close()

    if audit_mode:
        console.info("Use %ss to upload data return value %r", duration, content)

    if ret.status_code == 400:
        console.error("Bad request has been submitted for url %s, please report this to us, thank u.", url)
        raise DiscardDataForRequest()
    elif ret.status_code == 503:
        console.error("Agent server is unavailable. This can be a transient issue because of the server or our core"
                      " application being restarted. if this error continues, please report to us. thank u.")
        raise DiscardDataForRequest()
    elif ret.status_code == 502:
        console.error("Agent server error, our engineer has caution this error, thanks for your support")
        raise ServerIsUnavailable("service unavailable, get status code 502")
    elif ret.status_code != 200:
        console.warning("We got none 200 status code %s, this maybe some network/server error, if this error continues,"
                        "please report to us . thanks for your support. return content %s", ret.status_code, ret)
        raise DiscardDataForRequest()

    try:
        if six.PY3:
            content = content.decode('UTF-8')

        result = json.loads(content)
    except Exception as err:
        console.warning("Decoding data for Json error. please contact us for further investigation. %s", err)
        raise DiscardDataForRequest(str(sys.exc_info()[1]))
    else:
        # successful exchange with server
        if result["status"] == "success":
            return result["result"] if "result" in result else []

    console.info("get unexpected return,  there maybe some issue. %s", result)
    server_status = int(result["result"]["errorCode"])

    if server_status == 460:
        console.warning("Invalid license key, Please contact to networkbench for more help.")
        raise InvalidLicenseException("Invalid license key")
    elif server_status == 462:
        console.warning("Invalid data format, maybe something get wrong when json encoding.")
        raise DiscardDataForRequest(content)
    elif server_status == 461:
        console.warning("Invalid data token, if this error continues, please report to networkbench support for further"
                        " investigation")
        raise InvalidDataTokenException()
    elif server_status == -1:
        console.warning("Agent server error, our engineer has caution this error, thanks for your support.")
        raise ServerIsUnavailable()
    elif server_status == 470:
        console.info("Configuration is out of date, server configuration will be obtain again")
        raise OutOfDateConfigException()

    return []




"""this module is defined to detect the httplib2 module
"""

from tingyun.armoury.ammunition.external_tracker import wrap_external_trace


def detect_httplib2_http(module):
    """
    :param module:
    :return:
    """
    def http_url(instance, uri, method, *args, **kwargs):
        """
        :param instance:Http() instance
        :param uri:
        :param method:
        :param args:
        :param kwargs:
        :return:
        """
        return uri

    return wrap_external_trace(module, "Http.request", 'httplib2', http_url)


def detect_http_connect_with_timeout(module):
    """
    :param module:
    :return:
    """
    def http_with_timeout_url(instance, *args, **kwargs):
        """
        :param instance: HTTPConnectionWithTimeout instance
        :param args:
        :param kwargs:
        :return:
        """
        url = "http://%s" % instance.host
        if instance.port:
            url = 'http://%s:%s' % (instance.host, instance.port)

        return url

    wrap_external_trace(module, "HTTPConnectionWithTimeout.connect", 'httplib2', http_with_timeout_url)


def detect_https_connect_with_timeout(module):
    """
    :param module:
    :return:
    """
    def https_with_timeout_url(instance, *args, **kwargs):
        """
        :param instance: HTTPSConnectionWithTimeout instance
        :param args:
        :param kwargs:
        :return:
        """
        url = "http://%s" % instance.host
        if instance.port:
            url = 'http://%s:%s' % (instance.host, instance.port)

        return url

    wrap_external_trace(module, "HTTPSConnectionWithTimeout.connect", 'httplib2', https_with_timeout_url)




"""this module is implement some wrapper for trace the tronado module

"""

import logging
import weakref
import sys

from tingyun.logistics.object_name import callable_name
from tingyun.battlefield.proxy import proxy_instance
from tingyun.battlefield.tracer import Tracer
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.trigger.wsgi_tornado import process_environ
from tingyun.armoury.ammunition.function_tracker import FunctionTracker
from tingyun.logistics.basic_wrapper import FunctionWrapper


console = logging.getLogger(__name__)




"""
"""
import weakref

from tingyun.logistics.basic_wrapper import wrap_object, FunctionWrapper
from tingyun.battlefield.tracer import Tracer
from tingyun.battlefield.proxy import proxy_instance
from tingyun.armoury.trigger.wsgi_django import wsgi_wrapper_inline


def wsgi_application_wrapper(module, object_path, framework, version):
    """
    """
    wrap_object(module, object_path, wsgi_wrapper_inline, (framework, version))


def process_environ(environ, request):
    """the environ in the default environ may not include our need environment variable when apply to wsgi application
    :param environ: returned environment from the container request.
    :param request:
    :return:
    """
    # we need this the build the request information. but not included in the environment.
    ret = {}.update(environ)
    ret["REQUEST_URI"] = request.uri
    ret["HTTP_REFERER"] = request.headers.get('Referer')

    if not environ:
        ret['PATH_INFO'] = request.path
        ret['SCRIPT_NAME'] = ""
        ret['QUERY_STRING'] = request.query

    return ret


def wrap_wsgi_container(wrapped, framework, version):
    """
    :param wrapped:
    :param framework:
    :param version:
    :return:
    """
    def wrapper(wrapped, instance, request, *args, **kwargs):
        """
        """
        tracker = getattr(request, "_self_tracker", None)

        if not tracker:
            tracker = Tracer(proxy_instance(), process_environ(instance.environ, request))
            tracker.start_work()

            # we save the tracker info current request for next use.
            request._self_tracker = tracker
            tracker._self_current_request = weakref.ref(request)
        else:
            pass

    FunctionWrapper(wrapped, wrapper)


def wsgi_container_wrapper(module, object_path, framework, version):
    """
    :param module:
    :param object_path:
    :param framework:
    :param version:
    :return:
    """
    wrap_object(module, object_path, wrap_wsgi_container, (framework, version))




"""

"""

from tingyun.armoury.ammunition.external_tracker import wrap_external_trace


def detect(module):

    def url_request(request, method, url, *args, **kwargs):
        return url

    wrap_external_trace(module, 'RequestMethods.request', 'urllib3', url_request)




"""

"""
from tingyun.armoury.ammunition.external_tracker import wrap_external_trace


def detect(module):

    def urllib_url(url, *args, **kwargs):
        return url

    if hasattr(module, 'urlretrieve'):
        wrap_external_trace(module, 'urlretrieve', 'urllib', urllib_url)



import copy
import logging
import os
import re

from tingyun import get_version

console = logging.getLogger(__name__)
filter_pro = ["log_level_mapping", ]


class Settings(object):
    """
    """
    def __repr__(self):
        return repr(self.__dict__)

    def __iter__(self):
        """return the iteration for setting item
        """
        return iter(flatten_settings(self).items())


class ErrorCollectorSettings(Settings):
    pass


class ActionTracerSettings(Settings):
    pass


class TransactionTracerSettings(Settings):
    pass


_settings = Settings()
_settings.transaction_tracer = TransactionTracerSettings()
_settings.error_collector = ErrorCollectorSettings()
_settings.action_tracer = ActionTracerSettings()

# configure file
_settings.app_name = "Python App"
_settings.plugins = ''  # split with ','
_settings.license_key = "This is default license"
_settings.enabled = True
_settings.log_file = None
_settings.log_level = logging.INFO
_settings.audit_mode = False
_settings.ssl = True
_settings.daemon_debug = False
_settings.host = "redirect.networkbench.com"
_settings.port = None
_settings.tingyun_id_secret = ''

# python environment variable
_settings.config_file = None
_settings.enable_profile = True
_settings.max_profile_depth = 600

# internal use constance
_settings.shutdown_timeout = float(os.environ.get("TINGYUN_AGENT_SHUTDOWN_TIMEOUT", 2.5))
_settings.startup_timeout = float(os.environ.get("TINGYUN_AGENT_STARTUP_TIMEOUT", 0.0))
_settings.data_version = 1.0
_settings.agent_version = get_version()
_settings.data_report_timeout = 15.0
_settings.stack_trace_count = 30  # used to limit the depth of action tracer stack
_settings.explain_plan_count = 30  # used to limit the depth of sql explain tracer stack
_settings.action_tracer_nodes = 2000
_settings.slow_sql_count = 20
_settings.action_apdex = {}
_settings.web_action_uri_params_captured = {}
_settings.external_url_params_captured = {}

# server configuration & some limitation about the data
# set the default value for it.
_settings.action_tracer.enabled = True
_settings.action_tracer.action_threshold = 2 * 1000  # 2s
_settings.action_tracer.top_n = 40
_settings.action_tracer.stack_trace_threshold = 500  # 500ms

_settings.action_tracer.slow_sql = True
_settings.action_tracer.slow_sql_threshold = 500
_settings.action_tracer.log_sql = False
_settings.action_tracer.explain_enabled = True
_settings.action_tracer.explain_threshold = 500  # 500ms
_settings.action_tracer.record_sql = "obfuscated"

# transaction tracer settings
_settings.transaction_tracer.enabled = False

_settings.auto_action_naming = True
_settings.urls_captured = []
_settings.ignored_params = []

_settings.error_collector.enabled = True
_settings.error_collector.ignored_status_codes = []

_settings.apdex_t = 500
_settings.capture_params = False


_settings.log_level_mapping = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
}


def flatten_settings(settings):
    """get the iteration for setting item, include Settings object in Setting
    """

    def _flatten(settings_item, name, setting_object):
        for key, value in setting_object.__dict__.items():
            if key in filter_pro:
                console.debug("skip flatten property %s", key)
                continue

            if isinstance(value, Settings):
                if name:
                    _flatten(settings_item, '%s.%s' % (name, key), value)
                else:
                    _flatten(settings_item, key, value)
            else:
                if name:
                    settings_item['%s.%s' % (name, key)] = value
                else:
                    settings_item[key] = value

        return settings_item

    return _flatten({}, None, settings)


def get_upload_settings():
    """ construct the settings for upload to server
    :return: settings
    """
    dump_settings = {}
    
    settings = flatten_settings(_settings)
    for key in settings:
        dump_settings["nbs.%s" % key] = settings[key]

    return dump_settings


def apply_config_setting(settings_object, name, value):
    target = settings_object
    fields = name.split('.', 1)

    while len(fields) > 1:
        # for additional Settings from server
        if not hasattr(target, fields[0]):
            setattr(target, fields[0], Settings())
        target = getattr(target, fields[0])
        fields = fields[1].split('.', 1)

    # transfer the string ignored ignored_status_codes to list
    if fields[0] == "ignored_status_codes" and value:
        try:
            value = [int(v) for v in value.split(",")]
        except Exception as err:
            console.warning("got invalid ignore status code %s, errors %s", value, err)
    elif fields[0] == "ignored_status_codes" and not value:
        value = []

    # transfer the urls_captured to list
    # Warning in windows, the split with \n will get incorrect result
    if fields[0] == "urls_captured" and value:
        try:
            value = [re.compile(r"%s" % v) for v in value.split("\n")]
        except Exception as err:
            console.warning("compile re url failed, %s, %s", value, err)
    elif fields[0] == "urls_captured" and not value:
        value = []

    # transfer the ignored_params to list
    if fields[0] == "ignored_params" and value:
        value = [str(v) for v in value.split(",")]
    elif fields[0] == "ignored_params" and not value:
        value = []

    setattr(target, fields[0], value)


def uri_params_captured(original):
    """parse the web_action_uri_params_captured with protocal
    :param original:
    :return: formatted uri with param
    """
    uri_params = {}
    try:
        for uri_param in original.split('|'):
            parts = uri_param.split(",")
            uri = parts.pop(0).encode('utf-8')
            uri_params[uri] = str(parts)
    except Exception as err:
        console.error("Parse the uri occurred errors. %s", err)

    return uri_params


def merge_settings(server_side_config=None):
    """
    :param server_side_config: the config downloaded from server
    :return: the merged settings
    """
    settings_snapshot = copy.deepcopy(_settings)
    global_conf_name = ["applicationId", "enabled", "appSessionKey", "dataSentInterval", "apdex_t", "tingyunIdSecret"]

    if not server_side_config:
        console.warn("update server config failed %s, local settings will be used.", server_side_config)
        return settings_snapshot

    # pop the structured config
    server_conf = server_side_config.pop("config", {})

    # get the individual settings
    if 'actionApdex' in server_side_config:
        original_apdex = server_side_config.pop("actionApdex", {})
        urls_apdexs = dict((key, original_apdex[key]) for key in original_apdex)
        action_apdex_urls = dict((key.encode("utf8"), urls_apdexs[key]) for key in urls_apdexs if key)

        server_side_config.update({"action_apdex": action_apdex_urls})

    for name in global_conf_name:
        if name not in server_side_config:
            console.warning("Lost server configure %s", name)
            continue

        apply_config_setting(settings_snapshot, name, server_side_config.get(name, ""))

    # transfer specified params captured with same uri
    if 'nbs.web_action_uri_params_captured' in server_conf:
        web_uri_params_captured = server_conf.pop("nbs.web_action_uri_params_captured", '')
        settings_snapshot.web_action_uri_params_captured = uri_params_captured(web_uri_params_captured)

    if 'nbs.external_url_params_captured' in server_conf:
        external_uri_params_captured = server_conf.pop("nbs.external_url_params_captured", '')
        settings_snapshot.external_url_params_captured = uri_params_captured(external_uri_params_captured)

    for item in server_conf:
        value = server_conf[item]
        # drop the first part of name with point
        start_pos = str(item).find('.') + 1
        apply_config_setting(settings_snapshot, item[start_pos:], value)

    console.debug("return merged settings %s", settings_snapshot)
    return settings_snapshot


def global_settings():
    return _settings



"""
"""

import logging

from tingyun.armoury.trigger.wsgi_tornado import wsgi_application_wrapper, wsgi_container_wrapper
from tingyun.armoury.trigger.async_tornado import connection_on_headers_wrapper, connection_on_request_body_wrapper
from tingyun.armoury.trigger.async_tornado import http_request_finish_wrapper, connection_finish_wrapper


console = logging.getLogger(__name__)


def detect_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    import tornado
    version = tornado.version

    # new feature in tornado 4.x
    if hasattr(module, 'WSGIAdapter'):
        wsgi_application_wrapper(module.WSGIAdapter, "__call__", "Tornado", version)

    if hasattr(module, "WSGIApplication"):
        wsgi_application_wrapper(module.WSGIApplication, "__call__", "Tornado", version)

    if hasattr(module, "WSGIContainer"):
        wsgi_container_wrapper(module.WSGIApplication, "__call__", "Tornado", version)


# There is much different in tornado version 3.x and 4.x. So deal the two version separately
#
# detect the tornado 3.x

def detect_tornado3_main_process(module):
    """all of the data handled in HTTPConnection class, include build header/body/finish
    :param module:
    :return:
    """

    # first step for process
    if hasattr(module, 'HTTPConnection'):
        module.HTTPConnection._on_headers = connection_on_headers_wrapper(module.HTTPConnection._on_headers)

    # second step for process
    if hasattr(module, 'HTTPConnection'):
        module.HTTPConnection._on_request_body = connection_on_request_body_wrapper(module.HTTPConnection._on_request_body)

    # third step for process
    if hasattr(module, "HTTPRequest"):
        module.HTTPRequest.finish = http_request_finish_wrapper(module.HTTPRequest.finish)

    # fourth step for process
    if hasattr(module, "HTTPConnection"):
        module.HTTPConnection.finish = connection_finish_wrapper(module.HTTPConnection.finish)




"""define the trigger for the specified weapon

"""

import logging
import sys

from tingyun.logistics.basic_wrapper import wrap_object, FunctionWrapper
from tingyun.logistics.object_name import callable_name
from tingyun.armoury.ammunition.django_tracker import WSGIApplicationResponse
from tingyun.armoury.ammunition.function_tracker import FunctionTracker, function_trace_wrapper
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.battlefield.tracer import Tracer
from tingyun.battlefield.proxy import proxy_instance


console = logging.getLogger(__name__)


def wsgi_wrapper_inline(wrapped, framework, version=''):
    """ wrap the uwsgi application entrance
    :param wrapped: the method need to be wrapped
    :return:
    """

    # the wsgi.__call__ will passed two parameters with environ, and start_response
    def wrapper(wrapped, instance, args, kwargs):
        """More detail about the argument, read the wrapper doc
        """
        environ, start_response = args

        tracker = current_tracker()
        if tracker:
            return wrapped(environ, start_response)

        tracker = Tracer(proxy_instance(), environ, framework)
        tracker.start_work()

        # respect the wsgi protocol
        def _start_response(status, response_headers, *args):
            # deal the response data
            tracker.deal_response(status, response_headers, *args)
            _write = start_response(status, response_headers, *args)

            def write(data):
                ret = _write(data)
                return ret

            return write

        result = []
        try:
            tracker.set_tracker_name(callable_name(wrapped), priority=1)
            application = function_trace_wrapper(wrapped)
            with FunctionTracker(tracker, name='Application', group='Python.WSGI'):
                result = application(environ, _start_response)
        except:
            tracker.finish_work(*sys.exc_info())
            raise

        return WSGIApplicationResponse(tracker, result)

    return FunctionWrapper(wrapped, wrapper)


def wsgi_application_wrapper(module, object_path, *framework):
    """
    :param module:
    :param object_path:
    :return:
    """
    wrap_object(module, object_path, wsgi_wrapper_inline, *framework)


def wsgi_application_decorator(framework='undefined'):
    """
    :param framework: the framework of current use.
    :return:
    """
    framework = '' if framework is None else framework

    def decorator(wrapped):
        """
        :param wrapped:
        :return:
        """
        return wsgi_wrapper_inline(wrapped, framework)

    return decorator




"""this module is implement some wrapper for trace the tornado module

"""

import logging
import weakref
import sys

from tingyun.logistics.object_name import callable_name
from tingyun.battlefield.proxy import proxy_instance
from tingyun.battlefield.tracer import Tracer
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.trigger.wsgi_tornado import process_environ
from tingyun.armoury.ammunition.function_tracker import FunctionTracker
from tingyun.logistics.basic_wrapper import FunctionWrapper


console = logging.getLogger(__name__)


def connection_on_headers_wrapper(wrapped, framework="Tornado", version='0'):
    """This is entrance for handle the request. so we start our tracer from here.
    :param wrapped:
    :return:
    """
    def wrapper(wrapped, connection, args, kwargs):
        """At this wrapper. we only init some environment for agent.nothing need to be done.
        :param wrapped:wrapped func
        :param connection: connection instance for current request
        :param args:
        :param kwargs:
        :return:
        """
        # add some ensurence for potential error, if this occurred, that indicate some wrong with last trace.
        # Then we drop the last trace and ignore this trace now.
        tracer = current_tracker()
        if tracer:
            console.warning("Unexpected situation arise, but no side effect to use the agent. That's only indicate "
                            "some illogicality trace in tracer. if this continue, please report to us, thank u.")
            tracer.drop_tracker()
            return wrapped(*args, **kwargs)

        # check the request status.
        if connection.stream.closed() or connection._request_finished:
            console.info("stream closed(%s). or connection finished(%s).", connection.stream.closed(), connection._request_finished)
            return

        wrapped(*args, **kwargs)

        # Because of the asynchronous and single thread feature. we store the tracer in the request.
        # we use _self_ to prevent conflict to the wrapped func namespace.
        request = connection._request
        if not request or hasattr(request, '_self_tracer'):
            console.info("request(%s) or has _self_tracer(%s)", request, hasattr(request, '_self_tracer'))
            return

        tracer = Tracer(proxy_instance(), process_environ({}, request), framework)
        tracer.start_work()
        request._self_tracer = tracer
        tracer.set_tracker_name(callable_name(tracer))
        try:
            # between header and body maybe blocked. so we calculate the block time.
            # if some code blocked. the http server will deal the other request, so will should drop the tracer from
            # thread cache.
            request._self_request_finished = False
            tracer._self_request = weakref.ref(request)
            tracer.drop_tracker()
        except Exception as err:
            request._self_tracer = None
            tracer.__exit__(*sys.exc_info())
            console.exception("Tornado raise error in HTTPConnection on_headers. %s", err)
            raise

    return FunctionWrapper(wrapped, wrapper)


def connection_on_request_body_wrapper(wrapped):
    """
    :param wrapped:
    :return:
    """
    def wrapper(wrapped, connection, args, kwargs):
        """
        :param wrapped:
        :param connection:
        :param args:
        :param kwargs:
        :return:
        """
        request = connection._request
        if not hasattr(request, '_self_tracer') or not request._self_tracer:
            console.info("body do not get the tracer...")
            return wrapped(*args, **kwargs)

        # We can not the tracer at the tracert of the user application. so we should restore to the thread cache.
        tracer = request._self_tracer
        tracer.save_tracker()

        try:
            console.info("execute the body...")
            return wrapped(*args, **kwargs)
        except Exception as err:
            tracer.__exit__(*sys.exc_info())
            console.exception("Tornado raise error in HTTPConnection on_request_body. %s", err)
            raise

    return FunctionWrapper(wrapped, wrapper)


def http_request_finish_wrapper(wrapped):
    """
    :param wrapped:
    :return:
    """
    def wrapper(wrapped, request, args, kwargs):
        """
        :param wrapped:
        :param request:
        :param args:
        :param kwargs:
        :return:
        """
        tracer = getattr(request, '_self_tracer', None)
        if not tracer:
            return wrapped(*args, **kwargs)

        current_tracer = current_tracker()
        if current_tracer:
            console.info("got the tracker is not equal to current.")
            if tracer != current_tracer:
                try:
                    current_tracer.drop_transaction()
                    return wrapper(wrapped, request, args, kwargs)
                finally:
                    current_tracer.save_transaction()
            else:
                return wrapped(*args, **kwargs)

        tracer.save_tracker()

        try:
            if not request.connection.stream.writing():
                tracer.__exit__(None, None, None)
                request._self_tracer = None
            else:
                tracer.drop_transaction()
        except Exception as err:
            tracer.__exit__(*sys.exc_info())
            console.exception("Tornado raise error in HTTPRequest finish. %s", err)
            raise

    return FunctionWrapper(wrapped, wrapper)


def connection_finish_wrapper(wrapped):
    """
    :param wrapped:
    :return:
    """
    def wrapper(wrapped, connection, args, kwargs):
        """
        :param wrapped:
        :param connection:
        :param args:
        :param kwargs:
        :return:
        """
        request = connection._request
        tracer = current_tracker()

        if tracer:
            return wrapped(*args, **kwargs)
        else:
            if not hasattr(request, '_self_tracer'):
                return wrapped(*args, **kwargs)

            tracer = request._self_tracer
            tracer.save_tracker()
            console.info("get unfinished tracker.")

            try:
                result = wrapped(*args, **kwargs)
                tracer.__exit__(None, None, None)
            except Exception as err:
                tracer.__exit__(*sys.exc_info())
                console.exception("Tornado raise error in HTTPConnection finish. %s", err)
                raise
            finally:
                request._self_tracer = None

            return result

    return FunctionWrapper(wrapped, wrapper)



"""

"""


def defined_repertoire():
    """
    :return:
    """
    hookers = {
        "memcached": [
            {"target": "memcache", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_memcached'},
        ],

        # mysql db
        "mysql": [
            {"target": "MySQLdb", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        "pymysql": [
            {"target": "pymysql", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        "oursql": [
            {"target": "oursql", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],

        # oracle
        "oracle": [
            {"target": "cx_Oracle", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],

        # postgres SQL
        "postgresql": [
            {"target": "postgresql.interface.proboscis.dbapi2", 'hook_func': 'detect',
             'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        # postgres SQL
        "psycopg2": [
            {"target": "psycopg2", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        "psycopg2ct": [
            {"target": "psycopg2ct", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        "psycopg2cffi": [
            {"target": "psycopg2cffi", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],

        # ODBC A Python DB API 2 module for ODBC
        "pyodbc": [
            {"target": "pyodbc", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],

        # nosql mongodb
        "mongodb": [
            {"target": "pymongo.mongo_client", 'hook_func': 'detect_mongo_client',
             'hook_module': 'tingyun.armoury.database_mongo'},
            {"target": "pymongo.connection", 'hook_func': 'detect_connection',
             'hook_module': 'tingyun.armoury.database_mongo'},
            {"target": "pymongo.collection", 'hook_func': 'detect_collection',
             'hook_module': 'tingyun.armoury.database_mongo'},
        ],

        # nosql redis
        "redis": [
            {"target": "redis.connection", 'hook_func': 'detect_connection',
             'hook_module': 'tingyun.armoury.database_redis'},
            {"target": "redis.client", 'hook_func': 'detect_client_operation',
             'hook_module': 'tingyun.armoury.database_redis'},
        ],

        # external call
        "urllib": [
            {"target": "urllib", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.external_urllib'},
        ],
        "urllib2": [
            {"target": "urllib2", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.external_urllib2'},
        ],
        "urllib3": [
            {"target": "urllib3.request", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.external_urllib3'},
        ],
        "thrift": [
            {"target": "thrift.transport.TSocket", 'hook_func': 'detect_tsocket',
             'hook_module': 'tingyun.armoury.external_thrift'},
        ],

        "requests": [
            {"target": "requests.api", 'hook_func': 'detect_requests_request',
             'hook_module': 'tingyun.armoury.external_requests'},
            {"target": "requests.sessions", 'hook_func': 'detect_requests_sessions',
             'hook_module': 'tingyun.armoury.external_requests'},
        ],
        "httplib2": [
            {"target": "httplib2", 'hook_func': 'detect_httplib2_http',
             'hook_module': 'tingyun.armoury.external_httplib2'},
            {"target": "httplib2", 'hook_func': 'detect_http_connect_with_timeout',
             'hook_module': 'tingyun.armoury.external_httplib2'},
            {"target": "httplib2", 'hook_func': 'detect_https_connect_with_timeout',
             'hook_module': 'tingyun.armoury.external_httplib2'},
        ],

        # django, this weapon must not be removed.
        "django": [
            {"target": "django.core.handlers.base", 'hook_func': 'detect_middleware',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.core.handlers.wsgi", 'hook_func': 'detect_wsgi_entrance',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.core.urlresolvers", 'hook_func': 'detect_urlresolvers',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.views.generic.base", 'hook_func': 'detect_views_dispatch',
             'hook_module': 'tingyun.armoury.framework_django'},

            {"target": "django.template.loader_tags", 'hook_func': 'detect_template_block_render',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.template.base", 'hook_func': 'detect_django_template',
             'hook_module': 'tingyun.armoury.framework_django'},

            {"target": "django.http.multipartparser", 'hook_func': 'detect_http_multipartparser',
             'hook_module': 'tingyun.armoury.framework_django'},

            {"target": "django.core.mail", 'hook_func': 'detect_core_mail',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.core.mail.message", 'hook_func': 'detect_core_mail_message',
             'hook_module': 'tingyun.armoury.framework_django'},
        ],

        # flask, 0.6-1.0
        "flask": [
            {"target": "flask.app", 'hook_func': 'detect_wsgi_entrance',
             'hook_module': 'tingyun.armoury.framework_flask'},
            {"target": "flask.app", 'hook_func': 'detect_app_entrance',
             'hook_module': 'tingyun.armoury.framework_flask'},
            {"target": "flask.app", 'hook_func': 'detect_templates',
             'hook_module': 'tingyun.armoury.framework_flask'},
        ],

        # jinja2 2.3-2.8
        'jinja2': [
            {"target": "jinja2.loaders", 'hook_func': 'detect_template_loader',
             'hook_module': 'tingyun.armoury.template_jinja2'},
            {"target": "jinja2.environment", 'hook_func': 'detect_jinja2',
             'hook_module': 'tingyun.armoury.template_jinja2'},
            ],

        # "tornado": [
        #     {"target": "tornado.wsgi", 'hook_func': 'detect_wsgi_entrance',
        #      'hook_module': 'tingyun.armoury.framework_tornado'},
        #     {"target": "tornado.httpserver", 'hook_func': 'detect_tornado3_main_process',
        #      'hook_module': 'tingyun.armoury.framework_tornado'},
        # ]
    }

    return hookers




"""

"""

import logging

from tingyun.armoury.ammunition.timer import Timer
from tingyun.logistics.warehouse.external_node import ExternalNode
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.basic_wrapper import FunctionWrapper, wrap_object

console = logging.getLogger(__name__)


class ExternalTrace(Timer):
    """define the external trace common api.

    """
    def __init__(self, tracker, library, url, params=None):
        super(ExternalTrace, self).__init__(tracker)

        self.library = library
        self.url = url.split('?')[0]
        self.params = self.parse_request_params(url, params)

        signed_param = []
        for p in self.params:
            if tracker.settings and p in tracker.settings.external_url_params_captured.get(self.url, ""):
                signed_param.append("%s=%s&" % (p, self.params[p]))

        if 0 != len(signed_param):
            self.url = "%s?%s" % (self.url, ''.join(signed_param).rstrip('&'))

    def parse_request_params(self, url, post_data):
        """ Note: some url request lib passed in the post data in different way.
        so do not support post data params capture now.
        :param url:
        :param post_data:
        :return:
        """
        url_params = url.split("?")
        list_params = url_params[1] if 2 == len(url_params) else ""
        get_params = dict([tuple(p.split('=')) for p in list_params.split("&") if p])

        return get_params

    def create_node(self):
        return ExternalNode(library=self.library, url=self.url, children=self.children,
                            start_time=self.start_time, end_time=self.end_time, duration=self.duration,
                            exclusive=self.exclusive)

    def terminal_node(self):
        return True


def external_trace_wrapper(wrapped, library, url):

    def dynamic_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        if tracker is None:
            return wrapped(*args, **kwargs)

        _url = url
        if callable(url):
            if instance is not None:
                _url = url(instance, *args, **kwargs)
            else:
                _url = url(*args, **kwargs)

        with ExternalTrace(tracker, library, _url, kwargs):
            return wrapped(*args, **kwargs)

    def literal_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        with ExternalTrace(tracker, library, url, kwargs):
            return wrapped(*args, **kwargs)

    if callable(url):
        return FunctionWrapper(wrapped, dynamic_wrapper)

    return FunctionWrapper(wrapped, literal_wrapper)


def wrap_external_trace(module, object_path, library, url):
    wrap_object(module, object_path, external_trace_wrapper, (library, url))



import cgi
import logging
import threading

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from tingyun.armoury.ammunition.tracker import Tracker


console = logging.getLogger(__name__)


class Tracer(Tracker):
    """
    """
    def __init__(self, application, environ, framework="Python"):
        """
        """
        Tracker.__init__(self, application, environ, framework)

        self.get_thread_name()
        script_name = environ.get('SCRIPT_NAME', "")
        path_info = environ.get('PATH_INFO', "")
        self.request_uri = environ.get("REQUEST_URI", "")
        self.referer = environ.get("HTTP_REFERER", "")

        if self.request_uri:
            self.request_uri = urlparse.urlparse(self.request_uri)[2]
            
        if script_name or path_info:
            if not path_info:
                path = script_name
            elif not script_name:
                path = path_info
            else:
                path = script_name + path_info

            self.set_tracker_name(path, 'Uri', priority=1)

            if not self.request_uri:
                self.request_uri = path
        else:
            if self.request_uri:
                self.set_tracker_name(self.request_uri, 'Uri', priority=1)

        # get the param
        self._get_request_params(environ)

    def deal_response(self, status, response_headers, *args):
        """
        :param status:
        :param response_headers:
        :param args:
        :return:
        """
        try:
            self.http_status = int(status.split(' ')[0])
        except Exception as _:
            console.warning("get status code failed, status is %s", status)

    def _get_request_params(self, environ):
        """
        :param environ: cgi environment
        :return: referer
        """
        value = environ.get('QUERY_STRING', "")
        if value:
            params = {}

            try:
                params = urlparse.parse_qs(value, keep_blank_values=True)
            except Exception:
                try:
                    # method <parse_qs> only for backward compatibility,  method <parse_qs> new in python2.6
                    params = cgi.parse_qs(value, keep_blank_values=True)
                except Exception:
                    pass

            self.request_params.update(params)

    def get_thread_name(self):
        """
        :return: thread name
        """
        try:
            self.thread_name = threading.currentThread().getName()
        except Exception as err:
            console.info("Get thread name failed. %s", err)
            self.thread_name = "unknown"

        return self.thread_name

    def start_work(self):
        """
        :return:
        """
        self.__enter__()

    def finish_work(self, exc_type, exc_val, exc_tb):
        """
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.__exit__(exc_type, exc_val, exc_tb)




"""this module is defined to detect the requests module
"""

from tingyun.armoury.ammunition.external_tracker import wrap_external_trace


def detect_requests_request(module):
    """
    :param module:
    :return:
    """
    def request_url(method, url, *args, **kwargs):
        """
        """
        return url

    wrap_external_trace(module, 'request', 'requests', request_url)


def detect_requests_sessions(module):
    """
    :param module:
    :return:
    """
    def request_url(instance, method, url, *args, **kwargs):
        """
        """
        return url

    wrap_external_trace(module, 'Session.request', 'requests', request_url)




"""Define this module for basic armory for flask

"""

from tingyun.armoury.trigger.wsgi_django import wsgi_application_wrapper
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace
from tingyun.armoury.ammunition.flask_tracker import add_url_rule_wrapper, handle_exception_wrapper
from tingyun.logistics.basic_wrapper import wrap_function_wrapper


def detect_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    wsgi_application_wrapper(module.Flask, '__call__', ('flask', ))


def detect_app_entrance(module):
    """
    :param module:
    :return:
    """
    # trace the views function call metric
    wrap_function_wrapper(module.Flask, 'add_url_rule', add_url_rule_wrapper)
    wrap_function_wrapper(module.Flask, 'handle_exception', handle_exception_wrapper)

    # new feature form flask 0.7
    if hasattr(module.Flask, 'handle_user_exception'):
        wrap_function_wrapper(module.Flask, 'handle_user_exception', handle_exception_wrapper)

    # new feature form flask 0.7
    if hasattr(module.Flask, 'full_dispatch_request'):
        wrap_function_trace(module.Flask, 'full_dispatch_request')


def detect_templates(module):
    """
    :param module:
    :return:
    """
    wrap_function_trace(module, '_render')